@@ -4,22 +4,21 @@
  * radiusd.h	Structures, prototypes and global variables
  *		for the FreeRADIUS server.
  *
- * Version:	$Id: radiusd.h,v 1.139.2.1 2003/09/02 03:08:52 phampson Exp $
+ * Version:	$Id: radiusd.h,v 1.156.2.1 2004/07/05 15:38:22 aland Exp $
  *
  */
 #include "libradius.h"
 #include "radpaths.h"
 #include "conf.h"
-#include "missing.h"
 #include "conffile.h"
 
 #include <stdarg.h>
 
-#if HAVE_UNISTD_H
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 
-#if HAVE_PTHREAD_H
+#ifdef HAVE_PTHREAD_H
 #include	<pthread.h>
 typedef pthread_t child_pid_t;
 #define child_kill pthread_kill
@@ -28,12 +27,29 @@ typedef pid_t child_pid_t;
 #define child_kill kill
 #endif
 
+#ifdef HAVE_NETINET_IN_H
+#include <netinet/in.h>
+#endif
+
+#ifdef HAVE_ARPA_INET_H
+#include <arpa/inet.h>
+#endif
+
+#include "missing.h"
+
 #define NO_SUCH_CHILD_PID (child_pid_t) (0)
 
 #ifndef NDEBUG
 #define REQUEST_MAGIC (0xdeadbeef)
 #endif
 
+/*
+ *	See util.c
+ */
+typedef struct request_data_t request_data_t;
+
+#define REQUEST_DATA_REGEX (0xadbeef00)
+
 typedef struct auth_req {
 #ifndef NDEBUG
 	uint32_t		magic; /* for debugging only */
@@ -45,6 +61,7 @@ typedef struct auth_req {
 	VALUE_PAIR		*config_items;
 	VALUE_PAIR		*username;
 	VALUE_PAIR		*password;
+	request_data_t		*data;
 	char			secret[32];
 	child_pid_t    		child_pid;
 	time_t			timestamp;
@@ -57,8 +74,8 @@ typedef struct auth_req {
 	 *	config file and SIGHUPed?
 	 */
 	char			proxysecret[32];
-	int			proxy_is_replicate;
 	int			proxy_try_count;
+	int			proxy_outstanding;
 	time_t			proxy_next_try;
 
 	int                     simul_max;
@@ -67,13 +84,16 @@ typedef struct auth_req {
 
 	int			finished;
 	int			options; /* miscellanous options */
-	void			**container;
+	void			*container;
 } REQUEST;
 
 #define RAD_REQUEST_OPTION_NONE            (0)
 #define RAD_REQUEST_OPTION_LOGGED_CHILD    (1 << 0)
 #define RAD_REQUEST_OPTION_DELAYED_REJECT  (1 << 1)
 #define RAD_REQUEST_OPTION_DONT_CACHE      (1 << 2)
+#define RAD_REQUEST_OPTION_FAKE_REQUEST    (1 << 3)
+#define RAD_REQUEST_OPTION_REJECTED        (1 << 4)
+#define RAD_REQUEST_OPTION_PROXIED         (1 << 5)
 
 /*
  *  Function handler for requests.
@@ -130,6 +150,30 @@ typedef struct pair_list {
 	struct pair_list	*lastdefault;
 } PAIR_LIST;
 
+
+/*
+ *	Types of listeners.
+ */
+typedef enum RAD_LISTEN_TYPE {
+	RAD_LISTEN_NONE = 0,
+	RAD_LISTEN_AUTH,
+	RAD_LISTEN_ACCT,
+	RAD_LISTEN_PROXY
+} RAD_LISTEN_TYPE;
+
+
+/*
+ *	For listening on multiple IP's and ports.
+ */
+typedef struct rad_listen_t {
+	struct rad_listen_t *next; /* could be rbtree stuff */
+	uint32_t	ipaddr;
+	RAD_LISTEN_TYPE	type;
+	int		port;
+	int		fd;
+} rad_listen_t;
+
+
 typedef struct main_config_t {
 	struct main_config *next;
 	time_t		config_dead_time;
@@ -167,6 +211,7 @@ typedef struct main_config_t {
 	const char      *pid_file;
 	const char	*uid_name;
 	const char	*gid_name;
+	rad_listen_t	*listen;
 	CONF_SECTION	*config;
 	RADCLIENT	*clients;
 	REALM		*realms;
@@ -174,6 +219,7 @@ typedef struct main_config_t {
 
 #define DEBUG	if(debug_flag)log_debug
 #define DEBUG2  if (debug_flag > 1)log_debug
+#define DEBUG3  if (debug_flag > 2)log_debug
 
 #define SECONDS_PER_DAY		86400
 #define MAX_REQUEST_TIME	30
@@ -215,6 +261,8 @@ typedef enum radlog_dest_t {
 
 /*
  *	Global variables.
+ *
+ *	We really shouldn't have this many.
  */
 extern const char	*progname;
 extern int		debug_flag;
@@ -230,8 +278,6 @@ extern int		log_stripped_names;
 extern int		log_auth_detail;
 extern int		auth_port;
 extern int		acct_port;
-extern int		proxy_port;
-extern int		proxyfd;
 extern const char      *radiusd_version;
 
 /*
@@ -244,14 +290,16 @@ int		rad_accounting(REQUEST *);
 /* session.c */
 int		rad_check_ts(uint32_t nasaddr, unsigned int port, const char *user,
 			     const char *sessionid);
-int		session_zap(int fd, uint32_t nasaddr, unsigned int port, const char *user,
+int		session_zap(REQUEST *request, uint32_t nasaddr,
+			    unsigned int port, const char *user,
 			    const char *sessionid, uint32_t cliaddr,
-			    char proto, time_t t);
+			    char proto);
 
 /* radiusd.c */
+#ifndef _LIBRADIUS
 void		debug_pair(FILE *, VALUE_PAIR *);
+#endif
 int		log_err (char *);
-int		rad_process(REQUEST *, int);
 int		rad_respond(REQUEST *, RAD_REQUEST_FUNP fun);
 
 /* util.c */
@@ -261,6 +309,15 @@ int		rad_mkdir(char *directory, int mode);
 int		rad_checkfilename(const char *filename);
 void		*rad_malloc(size_t size); /* calls exit(1) on error! */
 void		xfree(const char *ptr);
+REQUEST		*request_alloc(void);
+REQUEST		*request_alloc_fake(REQUEST *oldreq);
+int		request_data_add(REQUEST *request,
+				 void *unique_ptr, int unique_int,
+				 void *opaque, void (*free_opaque)(void *));
+void		*request_data_get(REQUEST *request,
+				  void *unique_ptr, int unique_int);
+void		request_reject(REQUEST *request);
+void		rfc_clean(RADIUS_PACKET *packet);
 
 /* client.c */
 int		read_clients_file(const char *file);
@@ -312,6 +369,7 @@ int proxy_send(REQUEST *request);
 char	*auth_name(char *buf, size_t buflen, REQUEST *request, int do_cli);
 int		rad_authenticate (REQUEST *);
 int		rad_check_password(REQUEST *request);
+int		rad_postauth(REQUEST *);
 
 /* exec.c */
 int		radius_exec_program(const char *,  REQUEST *, int,
@@ -338,7 +396,7 @@ typedef int (*RADIUS_ESCAPE_STRING)(char *out, int outlen, const char *in);
 
 int            radius_xlat(char * out, int outlen, const char *fmt,
 			   REQUEST * request, RADIUS_ESCAPE_STRING func);
-typedef int (*RAD_XLAT_FUNC)(void *instance, REQUEST *, char *, char *, int, RADIUS_ESCAPE_STRING func);
+typedef int (*RAD_XLAT_FUNC)(void *instance, REQUEST *, char *, char *, size_t, RADIUS_ESCAPE_STRING func);
 int		xlat_register(const char *module, RAD_XLAT_FUNC func, void *instance);
 void		xlat_unregister(const char *module, RAD_XLAT_FUNC func);
 
@@ -346,6 +404,7 @@ void		xlat_unregister(const char *module, RAD_XLAT_FUNC func);
 /* threads.c */
 extern		int thread_pool_init(void);
 extern		int thread_pool_clean(time_t now);
+extern		int thread_pool_addrequest(REQUEST *, RAD_REQUEST_FUNP);
 extern		void rad_exec_init(void);
 extern		pid_t rad_fork(int exec_wait);
 extern		pid_t rad_waitpid(pid_t pid, int *status, int options);
