@@ -1,7 +1,7 @@
 /*
  * radius.c	Functions to send/receive radius packets.
  *
- * Version:	$Id: radius.c,v 1.99.2.5 2003/11/20 17:47:09 aland Exp $
+ * Version:	$Id: radius.c,v 1.125 2004/05/21 14:40:39 aland Exp $
  *
  *   This library is free software; you can redistribute it and/or
  *   modify it under the terms of the GNU Lesser General Public
@@ -17,17 +17,17 @@
  *   License along with this library; if not, write to the Free Software
  *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
  *
- * Copyright 2000  The FreeRADIUS server project
+ * Copyright 2000-2003  The FreeRADIUS server project
  */
 
-static const char rcsid[] = "$Id: radius.c,v 1.99.2.5 2003/11/20 17:47:09 aland Exp $";
+static const char rcsid[] = "$Id: radius.c,v 1.125 2004/05/21 14:40:39 aland Exp $";
 
 #include	"autoconf.h"
 #include	"md5.h"
 
 #include	<stdlib.h>
 
-#if HAVE_UNISTD_H
+#ifdef HAVE_UNISTD_H
 #include	<unistd.h>
 #endif
 
@@ -36,18 +36,21 @@ static const char rcsid[] = "$Id: radius.c,v 1.99.2.5 2003/11/20 17:47:09 aland
 #include	<ctype.h>
 
 #include	"libradius.h"
+#ifdef WITH_UDPFROMTO
+#include	"udpfromto.h"
+#endif
 
-#if HAVE_NETINET_IN_H
+#ifdef HAVE_NETINET_IN_H
 #include	<netinet/in.h>
 #endif
 
 #include	<sys/socket.h>
 
-#if HAVE_ARPA_INET_H
+#ifdef HAVE_ARPA_INET_H
 #include	<arpa/inet.h>
 #endif
 
-#if HAVE_MALLOC_H
+#ifdef HAVE_MALLOC_H
 #include	<malloc.h>
 #endif
 
@@ -158,6 +161,13 @@ int rad_send(RADIUS_PACKET *packet, const RADIUS_PACKET *original,
 	const char		*what;
 	uint8_t			ip_buffer[16];
 
+	/*
+	 *	Maybe it's a fake packet.  Don't send it.
+	 */
+	if (!packet || (packet->sockfd < 0)) {
+		return 0;
+	}
+
 	if ((packet->code > 0) && (packet->code < 52)) {
 		what = packet_codes[packet->code];
 	} else {
@@ -184,7 +194,7 @@ int rad_send(RADIUS_PACKET *packet, const RADIUS_PACKET *original,
 		   *	RADIUS packet size, by one attribute.
 		   */
 		  uint8_t		data[MAX_PACKET_LEN + 256];
-		  
+
 		  /*
 		   *	Use memory on the stack, until we know how
 		   *	large the packet will be.
@@ -196,20 +206,41 @@ int rad_send(RADIUS_PACKET *packet, const RADIUS_PACKET *original,
 		   */
 		  hdr->code = packet->code;
 		  hdr->id = packet->id;
-		  if ((packet->code == PW_ACCOUNTING_REQUEST) ||
-		      (packet->code == PW_DISCONNECT_REQUEST)) {
-			  memset(hdr->vector, 0, AUTH_VECTOR_LEN);
-		  } else {
-			  memcpy(hdr->vector, packet->vector, AUTH_VECTOR_LEN);
+
+		  /*
+		   *	Double-check some things based on packet code.
+		   */
+		  switch (packet->code) {
+		  case PW_AUTHENTICATION_ACK:
+		  case PW_AUTHENTICATION_REJECT:
+		  case PW_ACCESS_CHALLENGE:
+			  if (!original) {
+				  librad_log("ERROR: Cannot sign response packet without a request packet.");
+				  return -1;
+			  }
+			  break;
+
+			  /*
+			   *	These packet vectors start off as all zero.
+			   */
+		  case PW_ACCOUNTING_REQUEST:
+		  case PW_DISCONNECT_REQUEST:
+			  memset(packet->vector, 0, sizeof(packet->vector));
+			  break;
+
+		  default:
+			  break;
+
 		  }
+		  memcpy(hdr->vector, packet->vector, sizeof(hdr->vector));
 
 		  DEBUG("Sending %s of id %d to %s:%d\n",
 			what, packet->id,
 			ip_ntoa((char *)ip_buffer, packet->dst_ipaddr),
 			packet->dst_port);
-		  
+
 		  total_length = AUTH_HDR_LEN;
-		  
+
 		  /*
 		   *	Load up the configuration values for the user
 		   */
@@ -218,6 +249,9 @@ int rad_send(RADIUS_PACKET *packet, const RADIUS_PACKET *original,
 		  vendorpec = 0;
 		  vsa_length_ptr = NULL;
 
+		  /*
+		   *	Loop over the reply attributes for the packet.
+		   */
 		  for (reply = packet->vps; reply; reply = reply->next) {
 			  /*
 			   *	Ignore non-wire attributes
@@ -241,12 +275,10 @@ int rad_send(RADIUS_PACKET *packet, const RADIUS_PACKET *original,
 			  }
 
 			  /*
-			   *	Do stuff for Message-Authenticator
+			   *	Set the Message-Authenticator to the
+			   *	correct length and initial value.
 			   */
 			  if (reply->attribute == PW_MESSAGE_AUTHENTICATOR) {
-				  /*
-				   *  Set it to zero!
-				   */
 				  reply->length = AUTH_VECTOR_LEN;
 				  memset(reply->strvalue, 0, AUTH_VECTOR_LEN);
 				  msg_auth_offset = total_length;
@@ -296,7 +328,7 @@ int rad_send(RADIUS_PACKET *packet, const RADIUS_PACKET *original,
 			  if ((vendorcode == 0) &&
 			      ((vendorcode = VENDOR(reply->attribute)) != 0)) {
 				  vendorpec  = dict_vendorpec(vendorcode);
-				  
+
 				  /*
 				   *	This is a potentially bad error...
 				   *	we can't find the vendor ID!
@@ -329,9 +361,10 @@ int rad_send(RADIUS_PACKET *packet, const RADIUS_PACKET *original,
 				  ptr          += 4;
 
 				  /*
-				   *	Each USR attribute gets it's own
-				   *	VSA wrapper, so we re-set the
-				   *	vendor specific information.
+				   *	Each USR-style attribute gets
+				   *	it's own VSA wrapper, so we
+				   *	re-set the vendor specific
+				   *	information.
 				   */
 				  vendorcode = 0;
 				  vendorpec = 0;
@@ -342,35 +375,25 @@ int rad_send(RADIUS_PACKET *packet, const RADIUS_PACKET *original,
 				   *	All other attributes are as
 				   *	per the RFC spec.
 				   */
-
 				  *ptr++ = (reply->attribute & 0xFF);
 				  length_ptr = ptr;
 				  if (vsa_length_ptr) *vsa_length_ptr += 2;
 				  *ptr++ = 2;
 				  total_length += 2;
 			  }
-			  
+
 			  switch(reply->type) {
-				  
+
 				  /*
 				   *	Ascend binary attributes are
 				   *	stored internally in binary form.
 				   */
+			  case PW_TYPE_IFID:
+			  case PW_TYPE_IPV6ADDR:
+			  case PW_TYPE_IPV6PREFIX:
 			  case PW_TYPE_ABINARY:
 			  case PW_TYPE_STRING:
 			  case PW_TYPE_OCTETS:
-				  /*
-				   *  FIXME: HACK for non-updated dictionaries.
-				   *  REMOVE in a future release.
-				   */
-				  if ((strcmp(reply->name, "Ascend-Send-Secret") == 0) ||
-				      (strcmp(reply->name, "Ascend-Receive-Secret") == 0)) {
-					  reply->flags.encrypt = FLAG_ENCRYPT_ASCEND_SECRET;
-				  }
-				  if (reply->attribute == PW_USER_PASSWORD) {
-					  reply->flags.encrypt = FLAG_ENCRYPT_USER_PASSWORD;
-				  }
-
 				  /*
 				   *  Encrypt the various password styles
 				   */
@@ -386,10 +409,14 @@ int rad_send(RADIUS_PACKET *packet, const RADIUS_PACKET *original,
 				    break;
 
 				  case FLAG_ENCRYPT_TUNNEL_PASSWORD:
+					  if (!original) {
+						  librad_log("ERROR: No request packet, cannot encrypt Tunnel-Password attribute in the reply.");
+						  return -1;
+					  }
 					  rad_tunnel_pwencode(reply->strvalue,
 							      &(reply->length),
 							      secret,
-							      packet->vector);
+							      original->vector);
 					  break;
 
 
@@ -398,6 +425,7 @@ int rad_send(RADIUS_PACKET *packet, const RADIUS_PACKET *original,
 						      secret, reply->strvalue);
 					  memcpy(reply->strvalue, digest, AUTH_VECTOR_LEN );
 					  reply->length = AUTH_VECTOR_LEN;
+					  break;
 				  } /* switch over encryption flags */
 
 				  len = reply->length;
@@ -426,7 +454,7 @@ int rad_send(RADIUS_PACKET *packet, const RADIUS_PACKET *original,
 						  *ptr++ = 0x00;
 					  } /* else don't write a tag */
 				  } /* else the attribute doesn't have a tag */
-				 
+
 				  /*
 				   *	Ensure we don't go too far.
 				   *	The 'length' of the attribute
@@ -440,11 +468,11 @@ int rad_send(RADIUS_PACKET *packet, const RADIUS_PACKET *original,
 				  } else {
 					  allowed -= *length_ptr;
 				  }
-				  
+
 				  if (len > allowed) {
 					  len = allowed;
 				  }
-				  
+
 				  *length_ptr += len;
 				  if (vsa_length_ptr) *vsa_length_ptr += len;
 				  /*
@@ -457,7 +485,7 @@ int rad_send(RADIUS_PACKET *packet, const RADIUS_PACKET *original,
 				  ptr += reply->length;
 				  total_length += len;
 				  break;
-				  
+
 			  case PW_TYPE_INTEGER:
 			  case PW_TYPE_IPADDR:
 				  *length_ptr += 4;
@@ -531,79 +559,79 @@ int rad_send(RADIUS_PACKET *packet, const RADIUS_PACKET *original,
 		   *	and put it in the vector.
 		   */
 		  secretlen = strlen(secret);
-		  if (packet->code != PW_AUTHENTICATION_REQUEST &&
-		      packet->code != PW_STATUS_SERVER) {
-		    MD5_CTX	context;
-		      /*
-		       *	Set the Message-Authenticator attribute,
-		       *	BEFORE setting the reply authentication vector
-		       *	for CHALLENGE, ACCEPT and REJECT.
-		       */
-		      if (msg_auth_offset) {
-			      uint8_t calc_auth_vector[AUTH_VECTOR_LEN];
-
-			      switch (packet->code) {
-			      default:
-				break;
-				
-			      case PW_AUTHENTICATION_ACK:
-			      case PW_AUTHENTICATION_REJECT:
-			      case PW_ACCESS_CHALLENGE:
-				if (original) {
-				  memcpy(hdr->vector, original->vector, AUTH_VECTOR_LEN);
-				}
-				break;
-			      }
-
-			      memset(packet->data + msg_auth_offset + 2, 0,
-				     AUTH_VECTOR_LEN);
-			      lrad_hmac_md5(packet->data, packet->data_len,
-					    secret, secretlen, calc_auth_vector);
-			      memcpy(packet->data + msg_auth_offset + 2,
-				     calc_auth_vector, AUTH_VECTOR_LEN);
-			      memcpy(hdr->vector, packet->vector, AUTH_VECTOR_LEN);
-		      }
-
-		      MD5Init(&context);
-		      MD5Update(&context, packet->data, packet->data_len);
-		      MD5Update(&context, secret, strlen(secret));
-		      MD5Final(digest, &context);
-
-		      memcpy(hdr->vector, digest, AUTH_VECTOR_LEN);
-		      memcpy(packet->vector, digest, AUTH_VECTOR_LEN);
-		  }
 
 		  /*
-		   *	Set the Message-Authenticator attribute,
-		   *	AFTER setting the authentication vector
-		   *	only for ACCESS-REQUESTS
+		   *	If there's a Message-Authenticator, update it
+		   *	now, BEFORE updating the authentication vector.
 		   */
-		  else if (msg_auth_offset) {
+		  if (msg_auth_offset) {
 			  uint8_t calc_auth_vector[AUTH_VECTOR_LEN];
 
 			  switch (packet->code) {
 			  default:
-			    break;
-			    
+				  break;
+
 			  case PW_AUTHENTICATION_ACK:
 			  case PW_AUTHENTICATION_REJECT:
 			  case PW_ACCESS_CHALLENGE:
-			    if (original) {
-				    memcpy(hdr->vector, original->vector,
-					   AUTH_VECTOR_LEN);
-			    }
-			    break;
+				  /* this was checked above */
+				  memcpy(hdr->vector, original->vector,
+					 AUTH_VECTOR_LEN);
+				  break;
 			  }
 
+			  /*
+			   *	Set the authentication vector to zero,
+			   *	calculate the signature, and put it
+			   *	into the Message-Authenticator
+			   *	attribute.
+			   */
 			  memset(packet->data + msg_auth_offset + 2,
 				 0, AUTH_VECTOR_LEN);
 			  lrad_hmac_md5(packet->data, packet->data_len,
 					secret, secretlen, calc_auth_vector);
 			  memcpy(packet->data + msg_auth_offset + 2,
-				  calc_auth_vector, AUTH_VECTOR_LEN);
+				 calc_auth_vector, AUTH_VECTOR_LEN);
+
+			  /*
+			   *	Copy the original request vector back
+			   *	to the raw packet.
+			   */
 			  memcpy(hdr->vector, packet->vector, AUTH_VECTOR_LEN);
 		  }
 
+		  /*
+		   *	Switch over the packet code, deciding how to
+		   *	sign the packet.
+		   */
+		  switch (packet->code) {
+			  /*
+			   *	Request packets are not signed, bur
+			   *	have a random authentication vector.
+			   */
+		  case PW_AUTHENTICATION_REQUEST:
+		  case PW_STATUS_SERVER:
+			  break;
+
+			  /*
+			   *	Reply packets are signed with the
+			   *	authentication vector of the request.
+			   */
+		  default:
+		  	{
+				MD5_CTX	context;
+				MD5Init(&context);
+				MD5Update(&context, packet->data, packet->data_len);
+				MD5Update(&context, secret, strlen(secret));
+				MD5Final(digest, &context);
+
+				memcpy(hdr->vector, digest, AUTH_VECTOR_LEN);
+				memcpy(packet->vector, digest, AUTH_VECTOR_LEN);
+				break;
+			}
+		  } /* switch over packet codes */
+
+
 		  /*
 		   *	If packet->data points to data, then we print out
 		   *	the VP list again only for debugging.
@@ -612,13 +640,13 @@ int rad_send(RADIUS_PACKET *packet, const RADIUS_PACKET *original,
 	  	DEBUG("Re-sending %s of id %d to %s:%d\n", what, packet->id,
 		      ip_ntoa((char *)ip_buffer, packet->dst_ipaddr),
 		      packet->dst_port);
-		
+
 		for (reply = packet->vps; reply; reply = reply->next) {
 			/* FIXME: ignore attributes > 0xff */
 			debug_pair(reply);
 		}
 	}
-	
+
 	/*
 	 *	And send it on it's way.
 	 */
@@ -627,9 +655,21 @@ int rad_send(RADIUS_PACKET *packet, const RADIUS_PACKET *original,
 	sa->sin_family = AF_INET;
 	sa->sin_addr.s_addr = packet->dst_ipaddr;
 	sa->sin_port = htons(packet->dst_port);
-
+#ifndef WITH_UDPFROMTO
 	return sendto(packet->sockfd, packet->data, (int)packet->data_len, 0,
 		      (struct sockaddr *)&saremote, sizeof(struct sockaddr_in));
+#else
+	{
+		struct sockaddr_in salocal;
+		memset ((char *) &salocal, '\0', sizeof (salocal));
+		salocal.sin_family = AF_INET;
+		salocal.sin_addr.s_addr = packet->src_ipaddr;
+		
+		return sendfromto(packet->sockfd, packet->data, (int)packet->data_len, 0,
+				  (struct sockaddr *)&salocal,  sizeof(struct sockaddr_in),
+				  (struct sockaddr *)&saremote, sizeof(struct sockaddr_in));
+	}
+#endif
 }
 
 
@@ -659,7 +699,7 @@ static int calc_acctdigest(RADIUS_PACKET *packet, const char *secret)
 	 *	as the original MD5 sum (packet->vector).
 	 */
 	memset(packet->data + 4, 0, AUTH_VECTOR_LEN);
-	
+
 	/*
 	 *  MD5(packet + secret);
 	 */
@@ -732,22 +772,18 @@ RADIUS_PACKET *rad_recv(int fd)
 	socklen_t		salen;
 	u_short			len;
 	uint8_t			*attr;
-	uint8_t			*vendorattr;
 	int			count;
 	radius_packet_t		*hdr;
 	char			host_ipaddr[16];
 	int			seen_eap;
 	uint8_t			data[MAX_PACKET_LEN];
 	int			num_attributes;
-	uint32_t                vendorcode;
-	int			vendorlen;
-	
+
 	/*
 	 *	Allocate the new request data structure
 	 */
 	if ((packet = malloc(sizeof(RADIUS_PACKET))) == NULL) {
 		librad_log("out of memory");
-		errno = ENOMEM;
 		return NULL;
 	}
 	memset(packet, 0, sizeof(RADIUS_PACKET));
@@ -757,8 +793,22 @@ RADIUS_PACKET *rad_recv(int fd)
 	 */
 	salen = sizeof(saremote);
 	memset(&saremote, 0, sizeof(saremote));
+#ifndef WITH_UDPFROMTO
 	packet->data_len = recvfrom(fd, data, sizeof(data),
 		0, (struct sockaddr *)&saremote, &salen);
+	packet->dst_ipaddr = htonl(INADDR_ANY); /* i.e. unknown */
+#else
+	{
+		socklen_t		salen_local;
+		struct sockaddr_in	salocal;
+		salen_local = sizeof(salocal);
+		memset(&salocal, 0, sizeof(salocal));
+		packet->data_len = recvfromto(fd, data, sizeof(data), 0,
+					      (struct sockaddr *)&saremote, &salen,
+					      (struct sockaddr *)&salocal, &salen_local);
+		packet->dst_ipaddr = salocal.sin_addr.s_addr;
+	}
+#endif
 
 	/*
 	 *	Check for socket errors.
@@ -930,7 +980,7 @@ RADIUS_PACKET *rad_recv(int fd)
 			free(packet);
 			return NULL;
 		}
-		
+
 		/*
 		 *	Attributes are at LEAST as long as the ID & length
 		 *	fields.  Anything shorter is an invalid attribute.
@@ -964,46 +1014,32 @@ RADIUS_PACKET *rad_recv(int fd)
 			}
 			seen_eap |= PW_MESSAGE_AUTHENTICATOR;
 			break;
-			
+
 		case PW_VENDOR_SPECIFIC:
-			if (attr[1] < 6) {
+			if (attr[1] <= 6) {
 				librad_log("WARNING: Malformed RADIUS packet from host %s: Vendor-Specific has invalid length %d",
 					   ip_ntoa(host_ipaddr, packet->src_ipaddr),
 					   attr[1] - 2);
 				free(packet);
 				return NULL;
 			}
-			memcpy(&vendorcode, attr + 2, 4);
-			vendorcode = ntohl(vendorcode);
-			if (vendorcode == VENDORPEC_USR) {
-				if (attr[1] < 8){
-					librad_log("WARNING: Malformed RADIUS packet from host %s: USR attribute has invalid length %d",
-					   ip_ntoa(host_ipaddr, packet->src_ipaddr),
-					   attr[1] - 2);
-					free(packet);
-					return NULL;
-				}
-				break;
-			}
-			vendorlen = attr[1] - 6;
-			vendorattr = attr + 6;
-			while (vendorlen >= 2) {
-				if (vendorattr[1] < 2){
-					librad_log("WARNING: Malformed RADIUS packet from host %s: Vendor specific attribute has invalid length %d",
-					   ip_ntoa(host_ipaddr, packet->src_ipaddr),
-					   vendorattr[1] - 2);
-					free(packet);
-					return NULL;
-				}
-				vendorlen -= vendorattr[1];
-				vendorattr += vendorattr[1];
-			}
-			if (vendorlen != 0){
-				librad_log("WARNING: Malformed RADIUS packet from host %s: Vendor specific attributes do not exactly fill Vendor-Specific",
+
+			/*
+			 *	Don't allow VSA's with vendor zero.
+			 */
+			if ((attr[2] == 0) && (attr[3] == 0) &&
+			    (attr[4] == 0) && (attr[5] == 0)) {
+				librad_log("WARNING: Malformed RADIUS packet from host %s: Vendor-Specific has vendor ID of zero",
 					   ip_ntoa(host_ipaddr, packet->src_ipaddr));
 				free(packet);
 				return NULL;
 			}
+
+			/*
+			 *	Don't look at the contents of VSA's,
+			 *	too many vendors have non-standard
+			 *	formats.
+			 */
 			break;
 		}
 
@@ -1044,13 +1080,30 @@ RADIUS_PACKET *rad_recv(int fd)
 		return NULL;
 	}
 
+	/*
+	 * 	http://www.freeradius.org/rfc/rfc2869.html#EAP-Message
+	 *
+	 *	A packet with an EAP-Message attribute MUST also have
+	 *	a Message-Authenticator attribute.
+	 *
+	 *	A Message-Authenticator all by itself is OK, though.
+	 */
+	if (seen_eap &&
+	    (seen_eap != PW_MESSAGE_AUTHENTICATOR) &&
+	    (seen_eap != (PW_EAP_MESSAGE | PW_MESSAGE_AUTHENTICATOR))) {
+		librad_log("WARNING: Insecure packet from host %s:  Received EAP-Message with no Message-Authenticator.",
+			   ip_ntoa(host_ipaddr, packet->src_ipaddr));
+		free(packet);
+		return NULL;
+	}
+
 	if (librad_debug) {
 		if ((hdr->code > 0) && (hdr->code < 52)) {
 			printf("rad_recv: %s packet from host %s:%d",
 			       packet_codes[hdr->code],
 			       ip_ntoa(host_ipaddr, packet->src_ipaddr), packet->src_port);
 		} else {
-			printf("rad_recv: Packet from host %s:%d code=%d",	
+			printf("rad_recv: Packet from host %s:%d code=%d",
 			       ip_ntoa(host_ipaddr, packet->src_ipaddr), packet->src_port,
 			       hdr->code);
 		}
@@ -1085,7 +1138,6 @@ RADIUS_PACKET *rad_recv(int fd)
 int rad_decode(RADIUS_PACKET *packet, RADIUS_PACKET *original,
 	       const char *secret)
 {
-	DICT_ATTR		*attr;
 	uint32_t		lvalue;
 	uint32_t		vendorcode;
 	VALUE_PAIR		**tail;
@@ -1130,9 +1182,11 @@ int rad_decode(RADIUS_PACKET *packet, RADIUS_PACKET *original,
 			case PW_AUTHENTICATION_ACK:
 			case PW_AUTHENTICATION_REJECT:
 			case PW_ACCESS_CHALLENGE:
-			  if (original) {
-				  memcpy(packet->data + 4, original->vector, AUTH_VECTOR_LEN);
+			  if (!original) {
+				  librad_log("ERROR: Cannot validate Message-Authenticator in response packet without a request packet.");
+				  return -1;
 			  }
+			  memcpy(packet->data + 4, original->vector, AUTH_VECTOR_LEN);
 			  break;
 			}
 
@@ -1143,7 +1197,7 @@ int rad_decode(RADIUS_PACKET *packet, RADIUS_PACKET *original,
 				char buffer[32];
 				librad_log("Received packet from %s with invalid Message-Authenticator!  (Shared secret is incorrect.)",
 					   ip_ntoa(buffer, packet->src_ipaddr));
-				return 1;
+				return -1;
 			} /* else the message authenticator was good */
 
 			/*
@@ -1179,7 +1233,7 @@ int rad_decode(RADIUS_PACKET *packet, RADIUS_PACKET *original,
 				librad_log("Received Accounting-Request packet "
 				    "from %s with invalid signature!  (Shared secret is incorrect.)",
 				    ip_ntoa(buffer, packet->src_ipaddr));
-				return 1;
+				return -1;
 			}
 			break;
 
@@ -1191,11 +1245,12 @@ int rad_decode(RADIUS_PACKET *packet, RADIUS_PACKET *original,
 			if (rcode > 1) {
 				char buffer[32];
 				librad_log("Received %s packet "
-					   "from %s with invalid signature (err=%d)!  (Shared secret is incorrect.)",
+					   "from %s:%d with invalid signature (err=%d)!  (Shared secret is incorrect.)",
 					   packet_codes[packet->code],
 					   ip_ntoa(buffer, packet->src_ipaddr),
+					   packet->src_port,
 					   rcode);
-				return 1;
+				return -1;
 			}
 		  break;
 	}
@@ -1205,13 +1260,19 @@ int rad_decode(RADIUS_PACKET *packet, RADIUS_PACKET *original,
 	 */
 	ptr = hdr->data;
 	length = packet->data_len - AUTH_HDR_LEN;
-	packet->vps = NULL;
-	tail = &packet->vps;
+
+	/*
+	 *	There may be VP's already in the packet.  Don't
+	 *	destroy them.
+	 */
+	for (tail = &packet->vps; *tail != NULL; tail = &((*tail)->next)) {
+		/* nothing */
+	}
 
 	vendorcode = 0;
 	vendorlen  = 0;
 
-	while(length > 0) {
+	while (length > 0) {
 		if (vendorlen > 0) {
 			attribute = *ptr++ | (vendorcode << 16);
 			attrlen   = *ptr++;
@@ -1225,76 +1286,159 @@ int rad_decode(RADIUS_PACKET *packet, RADIUS_PACKET *original,
 
 		/*
 		 *	This could be a Vendor-Specific attribute.
-		 *
 		 */
 		if ((vendorlen <= 0) &&
-		    (attribute == PW_VENDOR_SPECIFIC) && 
-		    (attrlen > 6)) {
+		    (attribute == PW_VENDOR_SPECIFIC)) {
+			int	sublen;
+			uint8_t	*subptr;
+
+			/*
+			 *	attrlen was checked to be >= 6, in rad_recv
+			 */
+
 			memcpy(&lvalue, ptr, 4);
 			vendorcode = ntohl(lvalue);
-			if (vendorcode != 0) {
 
-				if (vendorcode == VENDORPEC_USR) {
-					ptr += 4;
-					memcpy(&lvalue, ptr, 4);
-					/*printf("received USR %04x\n", ntohl(lvalue));*/
-					attribute = (ntohl(lvalue) & 0xFFFF) |
-							(vendorcode << 16);
-					ptr += 4;
+			/*
+			 *	This is an implementation issue.
+			 *	We currently pack vendor into the upper
+			 *	16 bits of a 32-bit attribute number,
+			 *	so we can't handle vendor numbers larger
+			 *	than 16 bits.
+			 */
+			if (vendorcode > 65535) goto create_pair;
+
+			/*
+			 *	vendorcode was checked to be non-zero
+			 *	above, in rad_recv.
+			 */
+
+			/*
+			 *	First, check to see if the
+			 *	sub-attributes fill the VSA, as
+			 *	defined by the RFC.  If not, then it
+			 *	may be a USR-style VSA, or it may be a
+			 *	vendor who packs all of the
+			 *	information into one nonsense
+			 *	attribute
+			 */
+			subptr = ptr + 4;
+			sublen = attrlen - 4;
+
+			while (sublen > 0) {
+				if (subptr[1] < 2) { /* too short */
+					break;
+				}
+
+				if (subptr[1] > sublen) { /* too long */
+					break;
+				}
+
+				sublen -= subptr[1]; /* just right */
+				subptr += subptr[1];
+			}
+
+			/*
+			 *	If the attribute is RFC compatible,
+			 *	then allow it as an RFC style VSA.
+			 */
+			if (sublen == 0) {
+				ptr += 4;
+				vendorlen = attrlen - 4;
+				attribute = *ptr++ | (vendorcode << 16);
+				attrlen   = *ptr++;
+				attrlen -= 2;
+				length -= 6;
+
+				/*
+				 *	USR-style attributes are 4 octets,
+				 *	with the upper 2 octets being zero.
+				 *
+				 *	The upper octets may not be zero,
+				 *	but that then means we won't be
+				 *	able to pack the vendor & attribute
+				 *	into a 32-bit number, so we can't
+				 *	handle it.
+				 *
+				 *
+				 *	FIXME: Update the dictionaries so
+				 *	that we key off of per-attribute
+				 *	flags "4-octet", instead of hard
+				 *	coding USR here.  This will also
+				 *	let us send packets with other
+				 *	vendors having 4-octet attributes.
+				 */
+			} else if ((vendorcode == VENDORPEC_USR) &&
+				   ((ptr[4] == 0) && (ptr[5] == 0))) {
+				DICT_ATTR *da;
+
+				da = dict_attrbyvalue((vendorcode << 16) |
+						      (ptr[6] << 8) |
+						      ptr[7]);
+
+				/*
+				 *	See if it's in the dictionary.
+				 *	If so, it's a valid USR style
+				 *	attribute.  If not, it's not...
+				 *
+				 *	Don't touch 'attribute' until
+				 *	we know what to do!
+				 */
+				if (da != NULL) {
+					attribute = ((vendorcode << 16) |
+						     (ptr[6] << 8) |
+						     ptr[7]);
+					ptr += 8;
 					attrlen -= 8;
 					length -= 8;
-
-				} else {
-					ptr += 4;
-					vendorlen = attrlen - 4;
-					attribute = *ptr++ | (vendorcode << 16);
-					attrlen   = *ptr++;
-					attrlen -= 2;
-					length -= 6;
-				}
-			}
-			/*
-			 *  Else the vendor wasn't found...
-			 */
-		}
+				} /* else it's not in the dictionary */
+			} /* else it was a stupid vendor format */
+		} /* else it wasn't a VSA */
 
 		/*
-		 *	FIXME: should we us paircreate() ?
+		 *	Create the attribute, setting the default type
+		 *	to 'octects'.  If the type in the dictionary
+		 *	is different, then the dictionary type will
+		 *	over-ride this one.
 		 */
-		if ((pair = malloc(sizeof(VALUE_PAIR))) == NULL) {
+	create_pair:
+		if ((pair = paircreate(attribute, PW_TYPE_OCTETS)) == NULL) {
 			pairfree(&packet->vps);
 			librad_log("out of memory");
-			errno = ENOMEM;
 			return -1;
 		}
-		
-		memset(pair, 0, sizeof(VALUE_PAIR));
-		if ((attr = dict_attrbyvalue(attribute)) == NULL) {
-			snprintf(pair->name, sizeof(pair->name), "Attr-%d", attribute);
-			pair->type = PW_TYPE_OCTETS;
-		} else {
-			strcpy(pair->name, attr->name);
-			pair->type = attr->type;
-			pair->flags = attr->flags;
-		}
-		pair->attribute = attribute;
+
 		pair->length = attrlen;
 		pair->operator = T_OP_EQ;
 		pair->next = NULL;
-		
+
 		switch (pair->type) {
-			
-		case PW_TYPE_OCTETS:
-		case PW_TYPE_ABINARY:
+
+			/*
+			 *	The attribute may be zero length,
+			 *	or it may have a tag, and then no data...
+			 */
 		case PW_TYPE_STRING:
-			if (pair->flags.has_tag &&
-			    pair->type == PW_TYPE_STRING) {
+			if (pair->flags.has_tag) {
 				int offset = 0;
 
-				if ((pair->length > 0) && TAG_VALID(*ptr)) {
+				/*
+				 *	If there's sufficient room for
+				 *	a tag, and the tag looks valid,
+				 *	then use it.
+				 */
+				if ((pair->length > 0) &&
+				    TAG_VALID_ZERO(*ptr)) {
 					pair->flags.tag = *ptr;
 					pair->length--;
 					offset = 1;
+
+					/*
+					 *	If the leading tag
+					 *	isn't valid, then it's
+					 *	ignored for the tunnel
+					 *	password attribute.
+					 */
 				} else if (pair->flags.encrypt == FLAG_ENCRYPT_TUNNEL_PASSWORD) {
 					/*
 					 * from RFC2868 - 3.5.  Tunnel-Password
@@ -1312,28 +1456,22 @@ int rad_decode(RADIUS_PACKET *packet, RADIUS_PACKET *original,
 				}
 
 				/*
-				 *	pair->length may be zero here...
+				 *	pair->length MAY be zero here.
 				 */
 				memcpy(pair->strvalue, ptr + offset,
 				       pair->length);
 			} else {
+			  /*
+			   *	Ascend binary attributes never have a
+			   *	tag, and neither do the 'octets' type.
+			   */
+			case PW_TYPE_ABINARY:
+			case PW_TYPE_OCTETS:
 				/* attrlen always < MAX_STRING_LEN */
 				memcpy(pair->strvalue, ptr, attrlen);
 			        pair->flags.tag = 0;
 			}
 
-			/*
-			 *  FIXME: HACK for non-updated dictionaries.
-			 *  REMOVE in a future release.
-			 */
-			if ((strcmp(pair->name, "Ascend-Send-Secret") == 0) ||
-			    (strcmp(pair->name, "Ascend-Receive-Secret") == 0)) {
-				pair->flags.encrypt = FLAG_ENCRYPT_ASCEND_SECRET;
-			}
-			if (pair->attribute == PW_USER_PASSWORD) {
-				pair->flags.encrypt = FLAG_ENCRYPT_USER_PASSWORD;
-			}
-
 			/*
 			 *	Decrypt passwords here.
 			 */
@@ -1360,17 +1498,20 @@ int rad_decode(RADIUS_PACKET *packet, RADIUS_PACKET *original,
 				break;
 
 				/*
-				 *  Tunnel-Password
+				 *	Tunnel-Password's may go ONLY
+				 *	in response packets.
 				 */
 			case FLAG_ENCRYPT_TUNNEL_PASSWORD:
 				if (!original) {
 					librad_log("ERROR: Tunnel-Password attribute in request: Cannot decrypt it.");
 					return -1;
 				}
-			        rad_tunnel_pwdecode((char *)pair->strvalue,
-						    &pair->length, 
-						    secret,
-						    (char *)original->vector);
+				if (rad_tunnel_pwdecode(pair->strvalue,
+							&pair->length,
+							secret,
+							(char *)original->vector) < 0) {
+					return -1;
+				}
 				break;
 
 				/*
@@ -1391,7 +1532,7 @@ int rad_decode(RADIUS_PACKET *packet, RADIUS_PACKET *original,
 				break;
 			} /* switch over encryption flags */
 			break;	/* from octets/string/abinary */
-			
+
 		case PW_TYPE_INTEGER:
 		case PW_TYPE_DATE:
 		case PW_TYPE_IPADDR:
@@ -1412,7 +1553,7 @@ int rad_decode(RADIUS_PACKET *packet, RADIUS_PACKET *original,
 
       			memcpy(&lvalue, ptr, 4);
 
-			if (attr->type != PW_TYPE_IPADDR) {
+			if (pair->type != PW_TYPE_IPADDR) {
 				pair->lvalue = ntohl(lvalue);
 			} else {
 				 /*
@@ -1426,7 +1567,8 @@ int rad_decode(RADIUS_PACKET *packet, RADIUS_PACKET *original,
 			}
 
 			/*
-			 *  Only PW_TYPE_INTEGER should have tags.
+			 *	Tagged attributes of type integer have
+			 *	special treatment.
 			 */
 			if (pair->flags.has_tag &&
 			    pair->type == PW_TYPE_INTEGER) {
@@ -1434,7 +1576,11 @@ int rad_decode(RADIUS_PACKET *packet, RADIUS_PACKET *original,
 				pair->lvalue &= 0x00ffffff;
 			}
 
-			if (attr->type == PW_TYPE_INTEGER) {
+			/*
+			 *	Try to get the name for integer
+			 *	attributes.
+			 */
+			if (pair->type == PW_TYPE_INTEGER) {
 				DICT_VALUE *dval;
 				dval = dict_valbyattr(pair->attribute,
 						      pair->lvalue);
@@ -1445,15 +1591,57 @@ int rad_decode(RADIUS_PACKET *packet, RADIUS_PACKET *original,
 				}
 			}
 			break;
-			
+
+			/*
+			 *	IPv6 interface ID is 8 octets long.
+			 */
+		case PW_TYPE_IFID:
+			if (attrlen != 8)
+				pair->type = PW_TYPE_OCTETS;
+			memcpy(pair->strvalue, ptr, attrlen);
+			break;
+
+			/*
+			 *	IPv6 addresses are 16 octets long
+			 */
+		case PW_TYPE_IPV6ADDR:
+			if (attrlen != 16)
+				pair->type = PW_TYPE_OCTETS;
+			memcpy(pair->strvalue, ptr, attrlen);
+			break;
+
+			/*
+			 *	IPv6 prefixes are 2 to 18 octets long.
+			 *
+			 *	RFC 3162: The first octet is unused.
+			 *	The second is the length of the prefix
+			 *	the rest are the prefix data.
+			 *
+			 *	The prefix length can have value 0 to 128.
+			 */
+		case PW_TYPE_IPV6PREFIX:
+			if (attrlen < 2 || attrlen > 18)
+				pair->type = PW_TYPE_OCTETS;
+			if (attrlen >= 2) {
+				if (ptr[1] > 128) {
+					pair->type = PW_TYPE_OCTETS;
+				}
+				/*
+				 *	FIXME: double-check that
+				 *	(ptr[1] >> 3) matches attrlen + 2
+				 */
+			}
+			memcpy(pair->strvalue, ptr, attrlen);
+			break;
+
 		default:
 			DEBUG("    %s (Unknown Type %d)\n",
-			      attr->name,attr->type);
+			      pair->name, pair->type);
 			free(pair);
 			pair = NULL;
 			break;
 		}
-		
+
 		if (pair) {
 			debug_pair(pair);
 			*tail = pair;
@@ -1510,6 +1698,10 @@ int rad_pwencode(char *passwd, int *pwlen, const char *secret,
 		for (i = len; n > 0; n--, i++)
 			passwd[i] = 0;
 		len = *pwlen = i;
+
+	} else if (len == 0) {
+		memset(passwd, 0, AUTH_PASS_LEN);
+		*pwlen = len = AUTH_PASS_LEN;
 	}
 
 	/*
@@ -1532,7 +1724,7 @@ int rad_pwencode(char *passwd, int *pwlen, const char *secret,
 	 *	Length > AUTH_PASS_LEN, so we need to use the extended
 	 *	algorithm.
 	 */
-	for (n = 0; n < 128 && n <= (len - AUTH_PASS_LEN); n += AUTH_PASS_LEN) { 
+	for (n = 0; n < 128 && n <= (len - AUTH_PASS_LEN); n += AUTH_PASS_LEN) {
 		memcpy(buffer + secretlen, passwd + n, AUTH_PASS_LEN);
 		librad_md5_calc((u_char *)digest, buffer, secretlen + AUTH_PASS_LEN);
 		for (i = 0; i < AUTH_PASS_LEN; i++)
@@ -1581,7 +1773,7 @@ int rad_pwdecode(char *passwd, int pwlen, const char *secret,
 	 */
 	rlen = ((pwlen - 1) / AUTH_PASS_LEN) * AUTH_PASS_LEN;
 
-	for (n = rlen; n > 0; n -= AUTH_PASS_LEN ) { 
+	for (n = rlen; n > 0; n -= AUTH_PASS_LEN ) {
 		s = (n == AUTH_PASS_LEN) ? r : (passwd + n - AUTH_PASS_LEN);
 		memcpy(buffer + secretlen, s, AUTH_PASS_LEN);
 		librad_md5_calc((u_char *)digest, buffer, secretlen + AUTH_PASS_LEN);
@@ -1604,16 +1796,15 @@ static unsigned int salt_offset = 0;
  *      This is per RFC-2868 which adds a two char SALT to the initial intermediate
  *      value MD5 hash.
  */
-int rad_tunnel_pwencode(char *passwd, int *pwlen, const char *secret, const char *vector)
+int rad_tunnel_pwencode(char *passwd, int *pwlen, const char *secret,
+			const char *vector)
 {
 	uint8_t	buffer[AUTH_VECTOR_LEN + MAX_STRING_LEN + 3];
 	unsigned char	digest[AUTH_VECTOR_LEN];
 	char*   salt;
 	int	i, n, secretlen;
-	unsigned len;
-	
+	unsigned len, n2;
 
-	
 	len = *pwlen;
 
 	if (len > 127) len = 127;
@@ -1634,19 +1825,16 @@ int rad_tunnel_pwencode(char *passwd, int *pwlen, const char *secret, const char
 	/*
 	 *	Generate salt.  The RFC's say:
 	 *
-	 *	high bit of salt[0] must be set.
-	 *	each salt in a packet should be unique.
-	 *	they should be random
+	 *	The high bit of salt[0] must be set, each salt in a
+	 *	packet should be unique, and they should be random
 	 *
-	 *	So, we set the high bit,
-	 *	add in a counter,
-	 *	and then add in some CSPRNG data.
-	 *	should be OK..
+	 *	So, we set the high bit, add in a counter, and then
+	 *	add in some CSPRNG data.  should be OK..
 	 */
 	salt[0] = (0x80 | ( ((salt_offset++) & 0x0f) << 3) |
 		   (lrad_rand() & 0x07));
 	salt[1] = lrad_rand();
-	
+
 	/*
 	 *	Padd password to multiple of AUTH_PASS_LEN bytes.
 	 */
@@ -1658,89 +1846,138 @@ int rad_tunnel_pwencode(char *passwd, int *pwlen, const char *secret, const char
 	}
 	/* set new password length */
 	*pwlen = len + 2;
-	
+
 	/*
 	 *	Use the secret to setup the decryption digest
 	 */
 	secretlen = strlen(secret);
 	memcpy(buffer, secret, secretlen);
-	
-	for (n = 0; n < len; n+=AUTH_PASS_LEN) {
-		if (!n) {
+
+	for (n2 = 0; n2 < len; n2+=AUTH_PASS_LEN) {
+		if (!n2) {
 			memcpy(buffer + secretlen, vector, AUTH_VECTOR_LEN);
 			memcpy(buffer + secretlen + AUTH_VECTOR_LEN, salt, 2);
 			librad_md5_calc(digest, buffer, secretlen + AUTH_VECTOR_LEN + 2);
-		}
-		else {
-			memcpy(buffer + secretlen, passwd + n - AUTH_PASS_LEN, AUTH_PASS_LEN);
+		} else {
+			memcpy(buffer + secretlen, passwd + n2 - AUTH_PASS_LEN, AUTH_PASS_LEN);
 			librad_md5_calc(digest, buffer, secretlen + AUTH_PASS_LEN);
 		}
-		
-		for (i = 0; i < AUTH_PASS_LEN; i++)
-			passwd[i + n] ^= digest[i];
+
+		for (i = 0; i < AUTH_PASS_LEN; i++) {
+			passwd[i + n2] ^= digest[i];
+		}
 	}
-	passwd[n] = 0;
+	passwd[n2] = 0;
 	return 0;
 }
 
 /*
  *	Decode Tunnel-Password encrypted attributes.
  *
- *      Defined in RFC-2868, this adds a two char SALT to the initial intermediate
- *      value, to differentiate it from the above.
+ *      Defined in RFC-2868, this uses a two char SALT along with the
+ *      initial intermediate value, to differentiate it from the
+ *      above.
  */
-
-int rad_tunnel_pwdecode(char *passwd, int * pwlen, const char *secret, const char *vector)
+int rad_tunnel_pwdecode(uint8_t *passwd, int *pwlen, const char *secret,
+			const char *vector)
 {
-	uint8_t	buffer[AUTH_VECTOR_LEN + MAX_STRING_LEN + 3];
-	unsigned char	digest[AUTH_VECTOR_LEN];
-	char    salt[2];
-	int	i, n, ntimes, secretlen;
-	unsigned len;
-	
+	uint8_t		buffer[AUTH_VECTOR_LEN + MAX_STRING_LEN + 3];
+	uint8_t		digest[AUTH_VECTOR_LEN];
+	uint8_t		decrypted[MAX_STRING_LEN + 1];
+	int		secretlen;
+	unsigned	i, n, len;
+
 	len = *pwlen;
 
-	if(len < 3) {
-	  return len;
+	/*
+	 *	We need at least a salt.
+	 */
+	if (len < 2) {
+		librad_log("tunnel password is too short");
+		return -1;
 	}
-	salt[0] = passwd[0];
-	salt[1] = passwd[1];
 
-	passwd += 2;
-	len -= 2;
-	
+	/*
+	 *	There's a salt, but no password.  Or, there's a salt
+	 *	and a 'data_len' octet.  It's wrong, but at least we
+	 *	can figure out what it means: the password is empty.
+	 *
+	 *	Note that this means we ignore the 'data_len' field,
+	 *	if the attribute length tells us that there's no
+	 *	more data.  So the 'data_len' field may be wrong,
+	 *	but that's ok...
+	 */
+	if (len <= 3) {
+		passwd[0] = 0;
+		*pwlen = 0;
+		return 0;
+	}
+
+	len -= 2;		/* discount the salt */
+
 	/*
 	 *	Use the secret to setup the decryption digest
 	 */
 	secretlen = strlen(secret);
+
+	/*
+	 *	Set up the initial key:
+	 *
+	 *	 b(1) = MD5(secret + vector + salt)
+	 */
 	memcpy(buffer, secret, secretlen);
+	memcpy(buffer + secretlen, vector, AUTH_VECTOR_LEN);
+	memcpy(buffer + secretlen + AUTH_VECTOR_LEN, passwd, 2);
+	librad_md5_calc(digest, buffer, secretlen + AUTH_VECTOR_LEN + 2);
 
-	ntimes = (len-1)/AUTH_PASS_LEN;
-	do {
-		if(!ntimes){
-			memcpy(buffer + secretlen, vector, AUTH_VECTOR_LEN);
-			memcpy(buffer + secretlen + AUTH_VECTOR_LEN, salt, 2);
-			librad_md5_calc(digest, buffer, secretlen + AUTH_VECTOR_LEN + 2);
-		}
-		else {
-			memcpy(buffer + secretlen, passwd + AUTH_PASS_LEN * (ntimes - 1), AUTH_PASS_LEN);
-			librad_md5_calc(digest, buffer, secretlen + AUTH_PASS_LEN);
-		}
-		for ( i = 0, n = ntimes * AUTH_PASS_LEN; i < AUTH_PASS_LEN && (i + n) < len; i++)
-			passwd[i + n] ^= digest[i];
-	} while(ntimes--);
-	passwd[len] = '\0';
-
-	if (*(unsigned char*)passwd >= len) {
-		/* Pasword is broken, original password should be longer */
-		*pwlen = 2;
-		passwd[0]=passwd[1]=0;
-		return 0;
+	/*
+	 *	A quick check: decrypt the first octet of the password,
+	 *	which is the 'data_len' field.  Ensure it's sane.
+	 *
+	 *	'n' doesn't include the 'data_len' octet
+	 *	'len' does.
+	 */
+	n = passwd[2] ^ digest[0];
+	if (n >= len) {
+		librad_log("tunnel password is too long for the attribute");
+		return -1;
 	}
-	len = *pwlen = *passwd; /* restore original length */
-	for (n=0; n<len; n++) passwd[n-2]=passwd[n+1];
-	passwd[len-2] = 0;
-	return len;
+
+	/*
+	 *	Loop over the data, decrypting it, and generating
+	 *	the key for the next round of decryption.
+	 */
+	for (n = 0; n < len; n += AUTH_PASS_LEN) {
+		for (i = 0; i < AUTH_PASS_LEN; i++) {
+			decrypted[n + i] = passwd[n + i + 2] ^ digest[i];
+
+			/*
+			 *	Encrypted password may not be aligned
+			 *	on 16 octets, so we catch that here...
+			 */
+			if ((n + i) == len) break;
+		}
+
+		/*
+		 *	Update the digest, based on
+		 *
+		 *	b(n) = MD5(secret + cleartext(n-1)
+		 *
+		 *	but only if there's more data...
+		 */
+		memcpy(buffer + secretlen, passwd + n + 2, AUTH_PASS_LEN);
+		librad_md5_calc(digest, buffer, secretlen + AUTH_PASS_LEN);
+	}
+
+	/*
+	 *	We've already validated the length of the decrypted
+	 *	password.  Copy it back to the caller.
+	 */
+	memcpy(passwd, decrypted + 1, decrypted[0]);
+	passwd[decrypted[0]] = 0;
+	*pwlen = decrypted[0];
+
+	return decrypted[0];
 }
 
 /*
@@ -1791,7 +2028,7 @@ int rad_chap_encode(RADIUS_PACKET *packet, char *output, int id,
 		i += challenge->length;
 	} else {
 		memcpy(ptr, packet->vector, AUTH_VECTOR_LEN);
-		i += AUTH_VECTOR_LEN; 
+		i += AUTH_VECTOR_LEN;
 	}
 
 	*output = id;
