@@ -2,7 +2,7 @@
  * valuepair.c	Valuepair functions that are radiusd-specific
  *		and as such do not belong in the library.
  *
- * Version:	$Id: valuepair.c,v 1.49.2.1 2003/09/28 13:35:49 phampson Exp $
+ * Version:	$Id: valuepair.c,v 1.60.2.2 2004/07/14 21:01:02 aland Exp $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -22,7 +22,7 @@
  * Copyright 2000  Alan DeKok <aland@ox.org>
  */
 
-static const char rcsid[] = "$Id: valuepair.c,v 1.49.2.1 2003/09/28 13:35:49 phampson Exp $";
+static const char rcsid[] = "$Id: valuepair.c,v 1.60.2.2 2004/07/14 21:01:02 aland Exp $";
 
 #include "autoconf.h"
 #include "libradius.h"
@@ -31,7 +31,7 @@ static const char rcsid[] = "$Id: valuepair.c,v 1.49.2.1 2003/09/28 13:35:49 pha
 #include <stdlib.h>
 #include <string.h>
 
-#if HAVE_NETINET_IN_H
+#ifdef HAVE_NETINET_IN_H
 #	include <netinet/in.h>
 #endif
 
@@ -168,7 +168,7 @@ int paircompare_register(int attr, int compare_attr, RAD_COMPARE_FUNC fun, void
 
 	c = rad_malloc(sizeof(struct cmp));
 
-	if (compare_attr < 0) 
+	if (compare_attr < 0)
 		compare_attr = attr;
 	c->compare = fun;
 	c->attribute = attr;
@@ -241,6 +241,9 @@ int paircmp(REQUEST *req, VALUE_PAIR *request, VALUE_PAIR *check, VALUE_PAIR **r
 			case PW_CRYPT_PASSWORD:
 			case PW_AUTH_TYPE:
 			case PW_AUTZ_TYPE:
+			case PW_ACCT_TYPE:
+			case PW_SESSION_TYPE:
+			case PW_STRIP_USER_NAME:
 				continue;
 				break;
 
@@ -276,9 +279,24 @@ int paircmp(REQUEST *req, VALUE_PAIR *request, VALUE_PAIR *check, VALUE_PAIR **r
 		 *	Not found, it's not a match.
 		 */
 		if (auth_item == NULL) {
-			return -1;
+			/*
+			 *	Didn't find it.  If we were *trying*
+			 *	to not find it, then we succeeded.
+			 */
+			if (check_item->operator == T_OP_CMP_FALSE)
+				return 0;
+			else
+				return -1;
 		}
 
+		/*
+		 *	Else we found it, but we were trying to not
+		 *	find it, so we failed.
+		 */
+		if (check_item->operator == T_OP_CMP_FALSE)
+			return -1;
+
+
 		/*
 		 *	We've got to xlat the string before doing
 		 *	the comparison.
@@ -291,7 +309,7 @@ int paircmp(REQUEST *req, VALUE_PAIR *request, VALUE_PAIR *check, VALUE_PAIR **r
 			rcode = radius_xlat(buffer, sizeof(buffer),
 					    check_item->strvalue,
 					    req, NULL);
-			
+
 			/*
 			 *	Parse the string into a new value.
 			 */
@@ -326,7 +344,7 @@ int paircmp(REQUEST *req, VALUE_PAIR *request, VALUE_PAIR *check, VALUE_PAIR **r
 			case T_OP_GT:
 				if (compare <= 0) result = -1;
 				break;
-		    
+
 			case T_OP_LE:
 				if (compare > 0) result = -1;
 				break;
@@ -337,10 +355,70 @@ int paircmp(REQUEST *req, VALUE_PAIR *request, VALUE_PAIR *check, VALUE_PAIR **r
 
 #ifdef HAVE_REGEX_H
 			case T_OP_REG_EQ:
-				regcomp(&reg, (char *)check_item->strvalue, REG_EXTENDED|REG_NOSUB);
-				compare = regexec(&reg, (char *)auth_item->strvalue,
-						0, NULL, 0);
+			{
+				int i;
+				regmatch_t rxmatch[9];
+
+				/*
+				 *	Include substring matches.
+				 */
+				regcomp(&reg, (char *)check_item->strvalue,
+					REG_EXTENDED);
+				compare = regexec(&reg,
+						  (char *)auth_item->strvalue,
+						  16, rxmatch, 0);
 				regfree(&reg);
+
+				/*
+				 *	Add %{0}, %{1}, etc.
+				 */
+				for (i = 0; i <= 8; i++) {
+					char *p;
+					char buffer[sizeof(check_item->strvalue)];
+
+					/*
+					 *	Didn't match: delete old
+					 *	match, if it existed.
+					 */
+					if ((compare != 0) ||
+					    (rxmatch[i].rm_so == -1)) {
+						p = request_data_get(req, req,
+								     REQUEST_DATA_REGEX | i);
+						if (p) {
+							free(p);
+							continue;
+						}
+
+						/*
+						 *	No previous match
+						 *	to delete, stop.
+						 */
+						break;
+					}
+					
+					/*
+					 *	Copy substring into buffer.
+					 */
+					memcpy(buffer,
+					       auth_item->strvalue + rxmatch[i].rm_so,
+					       rxmatch[i].rm_eo - rxmatch[i].rm_so);
+					buffer[rxmatch[i].rm_eo - rxmatch[i].rm_so] = '\0';
+
+					/*
+					 *	Copy substring, and add it to
+					 *	the request.
+					 *
+					 *	Note that we don't check
+					 *	for out of memory, which is
+					 *	the only error we can get...
+					 */
+					p = strdup(buffer);
+					request_data_add(req,
+							 req,
+							 REQUEST_DATA_REGEX | i,
+							 p, free);
+				}
+			}				
 				if (compare != 0) result = -1;
 				break;
 
@@ -383,8 +461,12 @@ int simplepaircmp(REQUEST *req, VALUE_PAIR *first, VALUE_PAIR *second)
 /*
  *	Compare a Connect-Info and a Connect-Rate
  */
-static int connectcmp(void *instance, REQUEST *req, VALUE_PAIR *request, VALUE_PAIR *check,
-		VALUE_PAIR *check_pairs, VALUE_PAIR **reply_pairs)
+static int connectcmp(void *instance,
+		      REQUEST *req UNUSED,
+		      VALUE_PAIR *request,
+		      VALUE_PAIR *check,
+		      VALUE_PAIR *check_pairs,
+		      VALUE_PAIR **reply_pairs)
 {
 	int rate;
 
@@ -400,7 +482,8 @@ static int connectcmp(void *instance, REQUEST *req, VALUE_PAIR *request, VALUE_P
 /*
  *	Compare a portno with a range.
  */
-static int portcmp(void *instance, REQUEST *req, VALUE_PAIR *request, VALUE_PAIR *check,
+static int portcmp(void *instance,
+		   REQUEST *req UNUSED, VALUE_PAIR *request, VALUE_PAIR *check,
 	VALUE_PAIR *check_pairs, VALUE_PAIR **reply_pairs)
 {
 	char buf[MAX_STRING_LEN];
@@ -432,7 +515,7 @@ static int portcmp(void *instance, REQUEST *req, VALUE_PAIR *request, VALUE_PAIR
 			return 0;
 		}
 		s = strtok(NULL, ",");
-	} 
+	}
 
 	return -1;
 }
@@ -440,13 +523,15 @@ static int portcmp(void *instance, REQUEST *req, VALUE_PAIR *request, VALUE_PAIR
 /*
  *	Compare prefix/suffix.
  *
- *	If they compare: 
+ *	If they compare:
  *	- if PW_STRIP_USER_NAME is present in check_pairs,
  *	  strip the username of prefix/suffix.
  *	- if PW_STRIP_USER_NAME is not present in check_pairs,
  *	  add a PW_STRIPPED_USER_NAME to the request.
  */
-static int presufcmp(void *instance, REQUEST *req, VALUE_PAIR *request, VALUE_PAIR *check,
+static int presufcmp(void *instance,
+		     REQUEST *req UNUSED,
+		     VALUE_PAIR *request, VALUE_PAIR *check,
 	VALUE_PAIR *check_pairs, VALUE_PAIR **reply_pairs)
 {
 	VALUE_PAIR *vp;
@@ -454,7 +539,7 @@ static int presufcmp(void *instance, REQUEST *req, VALUE_PAIR *request, VALUE_PA
 	char rest[MAX_STRING_LEN];
 	int len, namelen;
 	int ret = -1;
-	
+
 	instance = instance;
 	reply_pairs = reply_pairs; /* shut the compiler up */
 
@@ -477,8 +562,7 @@ static int presufcmp(void *instance, REQUEST *req, VALUE_PAIR *request, VALUE_PA
 			ret = strcmp(name + namelen - len,
 					(char *)check->strvalue);
 			if (ret == 0 && rest) {
-				strncpy(rest, name, namelen - len);
-				rest[namelen - len] = 0;
+				strNcpy(rest, name, namelen - len + 1);
 			}
 			break;
 	}
@@ -510,13 +594,10 @@ static int presufcmp(void *instance, REQUEST *req, VALUE_PAIR *request, VALUE_PA
 
 /*
  *	Compare the current time to a range.
- *	Hmm... it would save work, and probably be better,
- *	if we were passed the REQUEST data structure, so we
- *	could use it's 'timestamp' element.  That way, we could
- *	do the comparison against when the packet came in, not now,
- *	and have one less system call to do.
  */
-static int timecmp(void *instance, REQUEST *req, VALUE_PAIR *request, VALUE_PAIR *check,
+static int timecmp(void *instance,
+		   REQUEST *req UNUSED,
+		   VALUE_PAIR *request, VALUE_PAIR *check,
 	VALUE_PAIR *check_pairs, VALUE_PAIR **reply_pairs)
 {
 	instance = instance;
@@ -524,7 +605,8 @@ static int timecmp(void *instance, REQUEST *req, VALUE_PAIR *request, VALUE_PAIR
 	check_pairs = check_pairs;
 	reply_pairs = reply_pairs;
 
-	if (timestr_match((char *)check->strvalue, time(NULL)) >= 0) {
+	if (timestr_match((char *)check->strvalue,
+			  req ? req->timestamp : time(NULL)) >= 0) {
 		return 0;
 	}
 	return -1;
@@ -539,8 +621,10 @@ static int timecmp(void *instance, REQUEST *req, VALUE_PAIR *request, VALUE_PAIR
  *	doing the lookup only ONCE, and storing the result
  *	in check->lvalue...
  */
-static int attrcmp(void *instance, REQUEST *req, VALUE_PAIR *request, VALUE_PAIR *check,
-	VALUE_PAIR *check_pairs, VALUE_PAIR **reply_pairs)
+static int attrcmp(void *instance,
+		   REQUEST *req UNUSED,
+		   VALUE_PAIR *request, VALUE_PAIR *check,
+		   VALUE_PAIR *check_pairs, VALUE_PAIR **reply_pairs)
 {
 	VALUE_PAIR *pair;
 	DICT_ATTR  *dict;
@@ -575,7 +659,8 @@ static int attrcmp(void *instance, REQUEST *req, VALUE_PAIR *request, VALUE_PAIR
 /*
  *	Compare the expiration date.
  */
-static int expirecmp(void *instance, REQUEST *req, VALUE_PAIR *request, VALUE_PAIR *check,
+static int expirecmp(void *instance, REQUEST *req UNUSED,
+		     VALUE_PAIR *request, VALUE_PAIR *check,
 		     VALUE_PAIR *check_pairs, VALUE_PAIR **reply_pairs)
 {
 	time_t now;
@@ -590,7 +675,7 @@ static int expirecmp(void *instance, REQUEST *req, VALUE_PAIR *request, VALUE_PA
 	 */
 	now = time(NULL);
 
-	if (now <= check->lvalue) {
+	if (now <= (signed)check->lvalue) {
 		return 0;
 	}
 
@@ -600,9 +685,11 @@ static int expirecmp(void *instance, REQUEST *req, VALUE_PAIR *request, VALUE_PA
 /*
  *	Compare the request packet type.
  */
-static int packetcmp(void *instance, REQUEST *req, VALUE_PAIR *request,
+static int packetcmp(void *instance UNUSED, REQUEST *req,
+		     VALUE_PAIR *request UNUSED,
 		     VALUE_PAIR *check,
-		     VALUE_PAIR *check_pairs, VALUE_PAIR **reply_pairs)
+		     VALUE_PAIR *check_pairs UNUSED,
+		     VALUE_PAIR **reply_pairs UNUSED)
 {
 	if (req->packet->code == check->lvalue) {
 		return 0;
@@ -614,9 +701,12 @@ static int packetcmp(void *instance, REQUEST *req, VALUE_PAIR *request,
 /*
  *	Compare the response packet type.
  */
-static int responsecmp(void *instance, REQUEST *req, VALUE_PAIR *request,
-		     VALUE_PAIR *check,
-		     VALUE_PAIR *check_pairs, VALUE_PAIR **reply_pairs)
+static int responsecmp(void *instance UNUSED,
+		       REQUEST *req,
+		       VALUE_PAIR *request UNUSED,
+		       VALUE_PAIR *check,
+		       VALUE_PAIR *check_pairs UNUSED,
+		       VALUE_PAIR **reply_pairs UNUSED)
 {
 	if (req->reply->code == check->lvalue) {
 		return 0;
@@ -687,7 +777,7 @@ void pairxlatmove(REQUEST *req, VALUE_PAIR **to, VALUE_PAIR **from)
 			rcode = radius_xlat(buffer, sizeof(buffer),
 					    i->strvalue,
 					    req, NULL);
-			
+
 			/*
 			 *	Parse the string into a new value.
 			 */
@@ -696,7 +786,7 @@ void pairxlatmove(REQUEST *req, VALUE_PAIR **to, VALUE_PAIR **from)
 
 		found = pairfind(*to, i->attribute);
 		switch (i->operator) {
-			
+
 			/*
 			 *  If a similar attribute is found,
 			 *  delete it.
@@ -707,7 +797,7 @@ void pairxlatmove(REQUEST *req, VALUE_PAIR **to, VALUE_PAIR **from)
 				    (strcmp((char *)found->strvalue,
 					    (char *)i->strvalue) == 0)){
 					pairdelete(to, found->attribute);
-					
+
 					/*
 					 *	'tailto' may have been
 					 *	deleted...
@@ -721,7 +811,7 @@ void pairxlatmove(REQUEST *req, VALUE_PAIR **to, VALUE_PAIR **from)
 			tailfrom = i;
 			continue;
 			break;
-			
+
 			/*
 			 *  Add it, if it's not already there.
 			 */
@@ -731,7 +821,7 @@ void pairxlatmove(REQUEST *req, VALUE_PAIR **to, VALUE_PAIR **from)
 				continue; /* with the loop */
 			}
 			break;
-			
+
 			/*
 			 *  If a similar attribute is found,
 			 *  replace it with the new one.  Otherwise,
@@ -747,14 +837,14 @@ void pairxlatmove(REQUEST *req, VALUE_PAIR **to, VALUE_PAIR **from)
 				continue;
 			}
 			break;
-			
+
 			/*
 			 *  FIXME: Add support for <=, >=, <, >
 			 *
 			 *  which will mean (for integers)
 			 *  'make the attribute the smaller, etc'
 			 */
-			
+
 			/*
 			 *  Add the new element to the list, even
 			 *  if similar ones already exist.
@@ -763,12 +853,12 @@ void pairxlatmove(REQUEST *req, VALUE_PAIR **to, VALUE_PAIR **from)
 		case T_OP_ADD:		/* += */
 			break;
 		}
-		
+
 		if (tailfrom)
 			tailfrom->next = next;
 		else
 			*from = next;
-		
+
 		/*
 		 *	If ALL of the 'to' attributes have been deleted,
 		 *	then ensure that the 'tail' is updated to point
