@@ -1,8 +1,8 @@
 /*
- * xlat.c	Translate strings.  This is the first version of xlat 
+ * xlat.c	Translate strings.  This is the first version of xlat
  * 		incorporated to RADIUS
  *
- * Version:	$Id: xlat.c,v 1.55 2003/03/18 05:50:54 aland Exp $
+ * Version:	$Id: xlat.c,v 1.72.2.2 2004/07/13 19:48:56 aland Exp $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -22,8 +22,8 @@
  * Copyright 2000  Alan DeKok <aland@ox.org>
  */
 
-static const char rcsid[] = 
-"$Id: xlat.c,v 1.55 2003/03/18 05:50:54 aland Exp $";
+static const char rcsid[] =
+"$Id: xlat.c,v 1.72.2.2 2004/07/13 19:48:56 aland Exp $";
 
 #include	"autoconf.h"
 #include	"libradius.h"
@@ -35,85 +35,33 @@ static const char rcsid[] =
 
 #include	"radiusd.h"
 
-struct xlat_cmp {
-	char module[MAX_STRING_LEN];
-	int length;
-	void *instance;
-	RAD_XLAT_FUNC do_xlat;
-	struct xlat_cmp *next;
-};
+#include	"rad_assert.h"
 
-static struct xlat_cmp *cmp = NULL;
+typedef struct xlat_t {
+	char		module[MAX_STRING_LEN];
+	int		length;
+	void		*instance;
+	RAD_XLAT_FUNC	do_xlat;
+	int		internal;	/* not allowed to re-define these */
+} xlat_t;
+
+static rbtree_t *xlat_root = NULL;
 
 /*
- *      Register an xlat function.
+ *	Define all xlat's in the structure.
  */
-int xlat_register(const char *module, RAD_XLAT_FUNC func, void *instance)
-{
-	struct xlat_cmp      *c;
+static const char *internal_xlat[] = {"check",
+				      "request",
+				      "reply",
+				      "proxy-request",
+				      "proxy-reply",
+				      NULL};
+static int xlat_inst[] = { 0, 1, 2, 3, 4, 5, 6, 7, 8 };	/* up to 8 for regex */
 
-	if (module == NULL || strlen(module) == 0){
-		DEBUG("xlat_register: Invalid module name");
-		return -1;
-	}
-
-	xlat_unregister(module, func);
-
-	c = rad_malloc(sizeof(struct xlat_cmp));
-
-	c->do_xlat = func;
-	strncpy(c->module, module, MAX_STRING_LEN);
-	c->length = strlen(c->module);
-	c->instance = instance;
-	c->next = cmp;
-	cmp = c;
-
-	return 0;
-}
 
 /*
- *      Unregister an xlat function.
+ *	Convert the value on a VALUE_PAIR to string
  */
-void xlat_unregister(const char *module, RAD_XLAT_FUNC func)
-{
-	struct xlat_cmp      *c, *last;
-
-	last = NULL;
-	for (c = cmp; c; c = c->next) {
-		if (strncmp(c->module,module,c->length) == 0 && c->do_xlat == func)
-			break;
-		last = c;
-	}
-
-	if (c == NULL) return;
-
-	if (last != NULL)
-		last->next = c->next;
-	else
-		cmp = c->next;
-
-	free(c);
-}
-
-/*
- * find the appropriate registered xlat function.
- */
-static struct xlat_cmp *find_xlat_func(const char *module)
-{
-	struct xlat_cmp *c;
-
-	for (c = cmp; c; c = c->next){
-		if (strncmp(c->module,module,c->length) == 0 && *(module+c->length) == ':')
-			break;
-	}
-
-	return c;
-}
-
-
-/*
-   Convert the value on a VALUE_PAIR to string
-*/
 static int valuepair2str(char * out,int outlen,VALUE_PAIR * pair,
 			 int type, RADIUS_ESCAPE_STRING func)
 {
@@ -143,21 +91,291 @@ static int valuepair2str(char * out,int outlen,VALUE_PAIR * pair,
 	return strlen(out);
 }
 
+
 /*
- *  Decode an attribute name into a string.
+ *	Dynamically translate for check:, request:, reply:, etc.
+ */
+static int xlat_packet(void *instance, REQUEST *request,
+		       char *fmt, char *out, size_t outlen,
+		       RADIUS_ESCAPE_STRING func)
+{
+	DICT_ATTR	*da;
+	VALUE_PAIR	*vp;
+	VALUE_PAIR	*vps = NULL;
+	RADIUS_PACKET	*packet = NULL;
+
+	switch (*(int*) instance) {
+	case 0:
+		vps = request->config_items;
+		break;
+
+	case 1:
+		vps = request->packet->vps;
+		packet = request->packet;
+		break;
+
+	case 2:
+		vps = request->reply->vps;
+		packet = request->reply;
+		break;
+
+	case 3:
+		if (request->proxy) vps = request->proxy->vps;
+		packet = request->proxy;
+		break;
+
+	case 4:
+		if (request->proxy_reply) vps = request->proxy_reply->vps;
+		packet = request->proxy_reply;
+		break;
+
+	default:		/* WTF? */
+		return 0;
+	}
+
+	/*
+	 *	The "format" string is the attribute name.
+	 */
+	da = dict_attrbyname(fmt);
+	if (!da) return 0;
+
+	vp = pairfind(vps, da->attr);
+	if (!vp) {
+		/*
+		 *	Some "magic" handlers, which are never in VP's, but
+		 *	which are in the packet.
+		 *
+		 *	FIXME: Add SRC/DST IP address!
+		 */
+		if (packet) {
+			switch (da->attr) {
+			case PW_PACKET_TYPE:
+			{
+				DICT_VALUE *dval;
+				
+				dval = dict_valbyattr(da->attr, packet->code);
+				if (dval) {
+					snprintf(out, outlen, "%s", dval->name);
+				} else {
+					snprintf(out, outlen, "%d", packet->code);
+				}
+				return strlen(out);
+			}
+			break;
+			
+			default:
+				break;
+			}
+		}
+
+		/*
+		 *	Not found, die.
+		 */
+		return 0;
+	}
+
+	if (!vps) return 0;	/* silently fail */
+
+	/*
+	 *	Convert the VP to a string, and return it.
+	 */
+	return valuepair2str(out, outlen, vp, da->type, func);
+}
+
+#ifdef HAVE_REGEX_H
+/*
+ *	Pull %{0} to %{8} out of the packet.
+ */
+static int xlat_regex(void *instance, REQUEST *request,
+		      char *fmt, char *out, size_t outlen,
+		      RADIUS_ESCAPE_STRING func)
+{
+	char *regex;
+
+	/*
+	 *	We cheat: fmt is "0" to "8", but those numbers
+	 *	are already in the "instance".
+	 */
+	fmt = fmt;		/* -Wunused */
+	func = func;		/* -Wunused FIXME: do escaping? */
+	
+	regex = request_data_get(request, request,
+				 REQUEST_DATA_REGEX | *(int *)instance);
+	if (!regex) return 0;
+
+	/*
+	 *	Copy UP TO "freespace" bytes, including
+	 *	a zero byte.
+	 */
+	strNcpy(out, regex, outlen);
+	free(regex); /* was strdup'd */
+	return strlen(out);
+}
+#endif				/* HAVE_REGEX_H */
+
+/*
+ *	Compare two xlat_t structs, based ONLY on the module name.
+ */
+static int xlat_cmp(const void *a, const void *b)
+{
+	if (((const xlat_t *)a)->length != ((const xlat_t *)b)->length) {
+		return ((const xlat_t *)a)->length - ((const xlat_t *)b)->length;
+	}
+
+	return memcmp(((const xlat_t *)a)->module,
+		      ((const xlat_t *)b)->module,
+		      ((const xlat_t *)a)->length);
+}
+
+
+/*
+ *	find the appropriate registered xlat function.
+ */
+static xlat_t *xlat_find(const char *module)
+{
+	char *p;
+	xlat_t my_xlat;
+
+	strNcpy(my_xlat.module, module, sizeof(my_xlat.module));
+
+	/*
+	 *	We get passed the WHOLE string, and all we want here
+	 *	is the first piece.
+	 */
+	p = strchr(my_xlat.module, ':');
+	if (p) *p = '\0';
+
+	my_xlat.length = strlen(my_xlat.module);
+
+	return rbtree_finddata(xlat_root, &my_xlat);
+}
+
+
+/*
+ *      Register an xlat function.
+ */
+int xlat_register(const char *module, RAD_XLAT_FUNC func, void *instance)
+{
+	xlat_t	*c;
+	xlat_t	my_xlat;
+
+	if ((module == NULL) || (strlen(module) == 0)) {
+		DEBUG("xlat_register: Invalid module name");
+		return -1;
+	}
+
+	/*
+	 *	First time around, build up the tree...
+	 *
+	 *	FIXME: This code should be hoisted out of this function,
+	 *	and into a global "initialization".  But it isn't critical...
+	 */
+	if (!xlat_root) {
+		int i;
+#ifdef HAVE_REGEX_H
+		char buffer[2];
+#endif
+
+		xlat_root = rbtree_create(xlat_cmp, free, 0);
+		if (!xlat_root) {
+			DEBUG("xlat_register: Failed to create tree.");
+			return -1;
+		}
+
+		/*
+		 *	Register the internal packet xlat's.
+		 */
+		for (i = 0; internal_xlat[i] != NULL; i++) {
+			xlat_register(internal_xlat[i], xlat_packet, &xlat_inst[i]);
+			c = xlat_find(internal_xlat[i]);
+			rad_assert(c != NULL);
+			c->internal = TRUE;
+		}
+
+#ifdef HAVE_REGEX_H
+		/*
+		 *	Register xlat's for regexes.
+		 */
+		buffer[1] = '\0';
+		for (i = 0; i <= 8; i++) {
+			buffer[0] = '0' + i;
+			xlat_register(buffer, xlat_regex, &xlat_inst[i]);
+			c = xlat_find(buffer);
+			rad_assert(c != NULL);
+			c->internal = TRUE;
+		}
+#endif /* HAVE_REGEX_H */
+	}
+
+	/*
+	 *	If it already exists, replace the instance.
+	 */
+	strNcpy(my_xlat.module, module, sizeof(my_xlat.module));
+	my_xlat.length = strlen(my_xlat.module);
+	c = rbtree_finddata(xlat_root, &my_xlat);
+	if (c) {
+		if (c->internal) {
+			DEBUG("xlat_register: Cannot re-define internal xlat");
+			return -1;
+		}
+
+		c->do_xlat = func;
+		c->instance = instance;
+		return 0;
+	}
+
+	/*
+	 *	Doesn't exist.  Create it.
+	 */
+	c = rad_malloc(sizeof(xlat_t));
+	memset(c, 0, sizeof(*c));
+
+	c->do_xlat = func;
+	strNcpy(c->module, module, sizeof(c->module));
+	c->length = strlen(c->module);
+	c->instance = instance;
+
+	rbtree_insert(xlat_root, c);
+
+	return 0;
+}
+
+/*
+ *      Unregister an xlat function.
+ *
+ *	We can only have one function to call per name, so the
+ *	passing of "func" here is extraneous.
+ */
+void xlat_unregister(const char *module, RAD_XLAT_FUNC func)
+{
+	rbnode_t	*node;
+	xlat_t		my_xlat;
+
+	func = func;		/* -Wunused */
+
+	strNcpy(my_xlat.module, module, sizeof(my_xlat.module));
+	my_xlat.length = strlen(my_xlat.module);
+
+	node = rbtree_find(xlat_root, &my_xlat);
+	if (!node) return;
+
+	rbtree_delete(xlat_root, node);
+}
+
+
+/*
+ *	Decode an attribute name into a string.
  */
 static void decode_attribute(const char **from, char **to, int freespace,
 			     int *open, REQUEST *request,
 			     RADIUS_ESCAPE_STRING func)
 {
-	DICT_ATTR *tmpda;
-	VALUE_PAIR *tmppair;
 	char attrname[256];
 	const char *p;
 	char *q, *pa;
-	int stop=0, found=0;
+	int stop=0, found=0, retlen=0;
 	int openbraces = *open;
-	struct xlat_cmp *c;
+	xlat_t *c;
 
 	p = *from;
 	q = *to;
@@ -165,10 +383,10 @@ static void decode_attribute(const char **from, char **to, int freespace,
 
 	*q = '\0';
 
-	/* 
-	 * Skip the '{' at the front of 'p' 
-	 * Increment open braces 
-	 */ 
+	/*
+	 * Skip the '{' at the front of 'p'
+	 * Increment open braces
+	 */
 	p++;
 	openbraces++;
 
@@ -189,7 +407,7 @@ static void decode_attribute(const char **from, char **to, int freespace,
 				openbraces++;
 				*pa++ = *p++;
 				break;
-				
+
 			case '}':
 				openbraces--;
 				if (openbraces == *open) {
@@ -222,63 +440,42 @@ static void decode_attribute(const char **from, char **to, int freespace,
 	*pa = '\0';
 
 	/*
-	 *	Find an attribute from the reply.
+	 *	Look up almost everything in the new tree of xlat
+	 *	functions.  this makes it a little quicker...
 	 */
-	if (strncasecmp(attrname,"reply:",6) == 0) {
-		if((tmpda = dict_attrbyname(&attrname[6])) && 
-				(tmppair = pairfind(request->reply->vps, tmpda->attr))) {
-			q += valuepair2str(q,freespace,tmppair,tmpda->type, func);
-			found = 1;
-		}
-
-		/*
-		 *	Find an attribute from the request.
-		 */
-	} else if (strncasecmp(attrname,"request:",8) == 0) {
-		if((tmpda = dict_attrbyname(&attrname[8])) && 
-				(tmppair = pairfind(request->packet->vps, tmpda->attr))) {
-			q += valuepair2str(q,freespace,tmppair,tmpda->type, func);
-			found = 1;
-		}
-
-		/*
-		 *	Find an attribute from the proxy reply.
-		 */
-	} else if ((strncasecmp(attrname,"proxy-reply:",12) == 0) &&
-		   (request->proxy_reply != NULL)) {
-		if((tmpda = dict_attrbyname(&attrname[12])) && 
-				(tmppair = pairfind(request->proxy_reply->vps, tmpda->attr))) {
-			q += valuepair2str(q,freespace,tmppair,tmpda->type, func);
-			found = 1;
-		}
-
-		/*
-		 *	Find a string from a registered function.
-		 */
-	} else if ((c = find_xlat_func(attrname)) != NULL) {
-		DEBUG("radius_xlat: Running registered xlat function of module %s for string \'%s\'",
-		      c->module, attrname+ c->length + 1);
-		q += c->do_xlat(c->instance, request, attrname+(c->length+1), q, freespace, func);
-		found = 1;
-
-		/*
-		 *	Nothing else, it MUST be a bare attribute name.
-		 */
-	} else {
-		tmpda = dict_attrbyname(attrname);
-		
-		if (tmpda) {
-			if ((tmppair = pairfind(request->packet->vps,tmpda->attr)) != NULL) {
-				q += valuepair2str(q,freespace,tmppair,tmpda->type, func);
-				found = 1;
-			} /* else silently fail */
+	if ((c = xlat_find(attrname)) != NULL) {
+		if (!c->internal) DEBUG("radius_xlat: Running registered xlat function of module %s for string \'%s\'",
+					c->module, attrname+ c->length + 1);
+		retlen = c->do_xlat(c->instance, request, attrname+(c->length+1), q, freespace, func);
+		/* If retlen is 0, treat it as not found */
+		if (retlen == 0) {
+			found = 0;
 		} else {
-			/*
-			 *	No attribute by that name, return an error.
-			 */
-			DEBUG2("WARNING: Attempt to use unknown xlat function or attribute in string %%{%s}", attrname);
+			found = 1;
+			q += retlen;
 		}
-	}
+
+		/*
+		 *	Not in the default xlat database.  Must be
+		 *	a bare attribute number.
+		 */
+	} else if ((retlen = xlat_packet(&xlat_inst[1], request, attrname,
+					 q, freespace, func)) > 0) {
+		found = 1;
+		q += retlen;
+
+		/*
+		 *	Look up the name, in order to get the correct
+		 *	debug message.
+		 */
+#ifndef NDEBUG
+	} else if (dict_attrbyname(attrname) == NULL) {
+		/*
+		 *	No attribute by that name, return an error.
+		 */
+		DEBUG2("WARNING: Attempt to use unknown xlat function, or non-existent attribute in string %%{%s}", attrname);
+#endif
+	} /* else the attribute is known, but not in the request */
 
 	/*
 	 * Skip to last '}' if attr is found
@@ -287,17 +484,20 @@ static void decode_attribute(const char **from, char **to, int freespace,
 	 */
 	if (found) {
 		while((*p != '\0') && (openbraces > 0)) {
+			/*
+			 *	Handle escapes outside of the loop.
+			 */
+			if (*p == '\\') {
+				p++;
+				if (!*p) break;
+				p++; /* get & ignore next character */
+				continue;
+			}
+
 			switch (*p) {
 			default:
 				break;
 
-				/*
-				 *  Ensure that escaped braces are allowed.
-				 */
-			case '\\':
-				p++; /* skip the escaped character */
-				break;
-
 				/*
 				 *  Bare brace
 				 */
@@ -367,6 +567,11 @@ int radius_xlat(char *out, int outlen, const char *fmt,
 	char tmpdt[40]; /* For temporary storing of dates */
 	int openbraces=0;
 
+	/*
+	 *	Catch bad modules.
+	 */
+	if (!fmt || !out || !request) return 0;
+
 	/*
 	 *  Ensure that we always have an escaping function.
 	 */
@@ -387,9 +592,9 @@ int radius_xlat(char *out, int outlen, const char *fmt,
 			/*
 			 * We check if we're inside an open brace.  If we are
 			 * then we assume this brace is NOT literal, but is
-			 * a closing brace and apply it 
+			 * a closing brace and apply it
 			 */
-			if((c == '}') && openbraces) {
+			if ((c == '}') && openbraces) {
 				openbraces--;
 				p++; /* skip it */
 				continue;
@@ -398,7 +603,15 @@ int radius_xlat(char *out, int outlen, const char *fmt,
 			continue;
 		}
 
-		if (*++p == '\0') break;
+		/*
+		 *	There's nothing after this character, copy
+		 *	the last '%' or "$' or '\\' over to the output
+		 *	buffer, and exit.
+		 */
+		if (*++p == '\0') {
+			*q++ = c;
+			break;
+		}
 
 		if (c == '\\') {
 			switch(*p) {
@@ -417,6 +630,13 @@ int radius_xlat(char *out, int outlen, const char *fmt,
 				break;
 			}
 			p++;
+
+			/*
+			 *	Hmmm... ${User-Name} is a synonym for
+			 *	%{User-Name}.
+			 *
+			 *	Why, exactly?
+			 */
 		} else if (c == '$') switch(*p) {
 			case '{': /* Attribute by Name */
 				decode_attribute(&p, &q, freespace, &openbraces, request, func);
@@ -458,7 +678,8 @@ int radius_xlat(char *out, int outlen, const char *fmt,
 				p++;
 				break;
 			case 'l': /* request timestamp */
-				snprintf(tmpdt, sizeof(tmpdt), "%lu",request->timestamp);
+				snprintf(tmpdt, sizeof(tmpdt), "%lu",
+					 (unsigned long) request->timestamp);
 				strNcpy(q,tmpdt,freespace);
 				q += strlen(q);
 				p++;
@@ -483,7 +704,7 @@ int radius_xlat(char *out, int outlen, const char *fmt,
 				p++;
 				break;
 			case 't': /* request timestamp */
-				ctime_r(&request->timestamp, q);
+				CTIME_R(&request->timestamp, q, freespace);
 				q += strlen(q);
 				p++;
 				break;
